# Copyright (c) 2022-2025 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'is_in',
    'get_filenames',
    'read_file',
    'read_stoich',
    'setup_network',
    'find_species',
]

import more_itertools
import numpy as np
import scipy.interpolate as si
import scipy.constants as sc

from .. import utils as u


def is_in(species):
    r"""
    Element-wise check whether species name exist in CEA database.

    Parameters
    ----------
    species: 1D iterable of strings
        Names of species to search in the database.

    Returns
    -------
    in_database: 1D bool array
        Flag whether each species is in the database.

    Examples
    --------
    >>> import chemcat.janaf as janaf
    >>> species = 'H2O (KOH)2 HO2 CO'.split()
    >>> in_janaf = janaf.is_in(species)
    >>> for spec, is_in in zip(species, in_janaf):
    >>>     print(f'{spec:6s}  {is_in}')
    H2O     True
    (KOH)2  True
    HO2     False
    CO      True
    """
    species = np.atleast_1d(species)
    janaf_names = [
        line.split()[0]
        for line in open(f'{u.ROOT}chemcat/data/janaf_conversion.txt', 'r')
    ]

    in_database = np.isin(species, janaf_names)
    return in_database


def get_filenames(species):
    """
    Convert species names to their respective JANAF file names.

    Parameters
    ----------
    species: String or 1D string iterable
        Species to search.

    Returns
    -------
    janaf_names: 1D string array
        Array of janaf filenames.  If a species is not found,
        return None in its place.

    Examples
    --------
    >>> import chemcat.janaf as janaf
    >>> species = 'H2O CH4 CO CO2 H2 e- H- H+ H2+ Na'.split()
    >>> janaf_species = janaf.get_filenames(species)
    >>> for mol, jname in zip(species, janaf_species):
    >>>     print(f'{mol:5}  {jname}')
    H2O    H-064.txt
    CH4    C-067.txt
    CO     C-093.txt
    CO2    C-095.txt
    H2     H-050.txt
    e-     D-020.txt
    H-     H-003.txt
    H+     H-002.txt
    H2+    H-051.txt
    Na     Na-005.txt
    """
    species = np.atleast_1d(species)

    janaf_dict = {}
    for line in open(f'{u.ROOT}chemcat/data/janaf_conversion.txt', 'r'):
        species_name, janaf_name = line.split()
        janaf_dict[species_name] = janaf_name

    janaf_names = [
        janaf_dict[molec] if molec in janaf_dict else None
        for molec in species
    ]
    return janaf_names


def read_file(janaf_file):
    """
    Read a JANAF file to extract tabulated thermal properties.

    Parameters
    ----------
    janaf_file: 1D string array
        A JANAF filename.

    Returns
    -------
    temps: 1D double array
        Tabulated JANAF temperatures (K).
    heat_capacity: 1D double array
        Tabulated JANAF heat capacity cp/R (unitless).
    gibbs_free_energy: 1D double array
        Tabulated JANAF Gibbs free energy G/RT (unitless).

    Examples
    --------
    >>> import chemcat.janaf as janaf
    >>> janaf_file = 'H-064.txt'  # Water
    >>> temps, heat, gibbs = janaf.read_file(janaf_file)
    >>> for i in range(5):
    >>>     print(f'{temps[i]:6.2f}  {heat[i]:.3f}  {gibbs[i]:.3f}')
    100.00  4.005  -317.133
    200.00  4.011  -168.505
    298.15  4.040  -120.263
    300.00  4.041  -119.662
    400.00  4.121  -95.583

    >>> temps, heat, gibbs = janaf.read_file(janaf_file)
    >>> for i in range(5):
    >>>     print(f'{temps[i]:6.2f}  {heat[i]:.3f}')
    298.15  2.500  -2.523
    300.00  2.500  -2.523
    350.00  2.500  -2.554
    400.00  2.500  -2.621
    450.00  2.500  -2.709
    """
    janaf_data = np.genfromtxt(
        f'{u.ROOT}chemcat/data/janaf/{janaf_file}',
        skip_header=3, usecols=(0,1,3,5), delimiter='\t',
        filling_values=np.nan,
        unpack=True,
    )
    temps = janaf_data[0]
    heat_capacity = janaf_data[1] / sc.R

    # https://janaf.nist.gov/pdf/JANAF-FourthEd-1998-1Vol1-Intro.pdf
    # Page 15
    df_H298 = janaf_data[3][temps==298.15]
    gibbs_free_energy = (-janaf_data[2] + df_H298*1000.0/temps) / sc.R

    idx_valid = np.isfinite(heat_capacity)
    return (
        temps[idx_valid],
        heat_capacity[idx_valid],
        gibbs_free_energy[idx_valid],
    )


def read_stoich(species=None, janaf_file=None, formula=None):
    """
    Get the stoichiometric data from the JANAF data base for the
    requested species.

    Parameters
    ----------
    species: String
        A species name (takes precedence over janaf_file argument).
    janaf_file: String
        A JANAF filename.
    formula: String
        A chemical formula in JANAF format (takes precedence over
        species and janaf_file arguments).

    Returns
    -------
    stoich: Dictionary
        Dictionary containing the stoichiometric values for the
        requested species. The dict's keys are the elements/electron
        names and their values are the respective stoich values.

    Examples
    --------
    >>> import chemcat.janaf as janaf
    >>> # From species name:
    >>> for species in 'C H2O e- H2+'.split():
    >>>     print(f'{species}:  {janaf.read_stoich(species)}')
    C:  {'C': 1.0}
    H2O:  {'H': 2.0, 'O': 1.0}
    e-:  {'e': 1.0}
    H2+:  {'e': -1, 'H': 2.0}

    >>> # From JANAF filename:
    >>> print(janaf.read_stoich(janaf_file='H-064.txt'))
    {'H': 2.0, 'O': 1.0}

    >>> # Or directly from the chemical formula:
    >>> print(janaf.read_stoich(formula='H3O1+'))
    {'e': -1, 'H': 3.0, 'O': 1.0}
    """
    # Get chemical formula (JANAF format):
    if formula is None and species is not None:
        janaf_file = get_filenames(species)[0]
    if formula is None and janaf_file is not None:
        with open(f'{u.ROOT}chemcat/data/janaf/{janaf_file}', 'r') as f:
            header = f.readline()
        formula = header.split('\t')[-1]
        formula = formula[0:formula.index('(')]

    if '-' in formula:
        stoich = {'e': 1}
    elif '+' in formula:
        stoich = {'e': -1}
    else:
        stoich = {}

    previous_type = formula[0].isalpha()
    word = ''
    groups = []
    for letter in formula.replace('-','').replace('+',''):
        if letter.isalpha() != previous_type:
            groups.append(word)
            word = ''
        word += letter
        previous_type = letter.isalpha()
    groups.append(word)
    for e, num in more_itertools.chunked(groups,2):
        stoich[e] = float(num)
    return stoich


def setup_network(input_species):
    r"""
    Extract JANAF thermal data for a requested chemical network.

    Parameters
    ----------
    species: 1D string iterable
        Species to search in the JANAF data base.

    Returns
    -------
    species: 1D string array
        Species found in the JANAF database (might differ from input_species).
    heat_capacity_splines: 1D list of numpy splines
        Splines sampling the species' heat capacity/R.
    gibbs_free_energy: 1D list of callable objects
        Functions that return the species's Gibbs free energy, G/RT.
    stoich_data: List of Dictionaries
        Stoichiometric data (as dictionary of element-value pairs) for
        a list of species.

    Examples
    --------
    >>> import chemcat.janaf as janaf

    >>> molecules = 'H2O CH4 CO CO2 NH3 N2 H2 HCN OH H He C N O'.split()
    >>> species, cp_funcs, gibbs_funcs, stoich_data = \
    >>>     janaf.setup_network(molecules)

    >>> for spec, stoich in zip(species, stoich_data):
    >>>     print(f'{spec:3s}:  {stoich}')
    H2O:  {'H': 2.0, 'O': 1.0}
    CH4:  {'C': 1.0, 'H': 4.0}
    CO :  {'C': 1.0, 'O': 1.0}
    CO2:  {'C': 1.0, 'O': 2.0}
    NH3:  {'H': 3.0, 'N': 1.0}
    N2 :  {'N': 2.0}
    H2 :  {'H': 2.0}
    HCN:  {'C': 1.0, 'H': 1.0, 'N': 1.0}
    OH :  {'H': 1.0, 'O': 1.0}
    H  :  {'H': 1.0}
    He :  {'He': 1.0}
    C  :  {'C': 1.0}
    N  :  {'N': 1.0}
    O  :  {'O': 1.0}
    """
    # Find which species exists in data base:
    janaf_species = get_filenames(input_species)
    nspecies = len(input_species)
    idx_missing = np.array([janaf is None for janaf in janaf_species])
    if np.any(idx_missing):
        missing_species = np.array(input_species)[idx_missing]
        print(f'These input species were not found:\n  {missing_species}')

    species = np.array(input_species)[~idx_missing]
    janaf_files = np.array(janaf_species)[~idx_missing]

    nspecies = len(species)
    heat_capacity = []
    gibbs_free_energy = []
    stoich_data = []
    for i in range(nspecies):
        janaf = janaf_files[i]
        janaf_data = read_file(janaf)
        temp = janaf_data[0]
        heat = janaf_data[1]
        gibbs = janaf_data[2]

        h_interp = si.interp1d(temp, heat, fill_value='extrapolate')
        g_interp = si.interp1d(
            temp, gibbs, fill_value='extrapolate', kind='cubic',
        )
        heat_capacity.append(h_interp)
        gibbs_free_energy.append(g_interp)
        stoich_data.append(read_stoich(janaf_file=janaf))

    return (
        species,
        heat_capacity,
        gibbs_free_energy,
        stoich_data,
    )


def find_species(elements, charge='neutral', num_atoms=None, state='gas'):
    """
    Find all JANAF species that contain the specified properties
    (elements, charge, state).

    Parameters
    ----------
    elements: Dict or 1D string iterable
        Either:
        - A list of elements that must be present in the species, or
        - A dictionary of elements and their stoichiometric values.
    charge: String
        If 'neutral', limit the output only to neutrally charged species.
        If 'ion', limit the output only to charged species.
        Else, do not limit output.
    num_atoms: Integer
        Limit the number of atoms to the requested value.
    state: String
        If 'gas', limit the output to gaseous species.

    Returns
    -------
    species: 1D string array
        List of all species containing the required elements.

    Examples
    --------
    >>> import chemcat.janaf as janaf

    >>> # Get all sodium-bearing species:
    >>> salts = janaf.find_species(['Na'])
    >>> print(salts)
    ['LiONa' 'Na2' 'Na2SO4' 'NaAlF4' 'NaBO2' '(NaBr)2' 'NaBr' '(NaCl)2' 'NaCl'
     '(NaCN)2' 'NaCN' '(NaF)2' 'NaF' 'Na' 'NaH' 'NaO' '(NaOH)2' 'NaOH']

    >>> # Get species containing exactly two Na atoms:
    >>> species = janaf.find_species({'Na':2})
    >>> print(species)
    ['Na2' 'Na2SO4' '(NaBr)2' '(NaCl)2' '(NaCN)2' '(NaF)2' '(NaOH)2']

    >>> # Species containing exactly two Na atoms and any amount of oxygen:
    >>> species = janaf.find_species({'Na':2, 'O':None})
    >>> print(species)
    ['Na2SO4' '(NaOH)2']

    >>> # Get all species containing sodium and oxygen (any amount):
    >>> species = janaf.find_species(['Na', 'O'])
    >>> print(species)
    ['LiONa' 'Na2SO4' 'NaBO2' 'NaO' '(NaOH)2' 'NaOH']

    >>> # Get all hydrogen-ion species:
    >>> H_ions = janaf.find_species(['H'], charge='ion')
    >>> print(H_ions)
    ['AlOH-' 'AlOH+' 'BaOH+' 'BeH+' 'BeOH+' 'CaOH+' 'CH+' 'CsOH+' 'H2-' 'H2+'
     'H3O+' 'HBO-' 'HBO+' 'HBS+' 'HCO+' 'HD-' 'HD+' 'H-' 'H+' 'KOH+' 'LiOH+'
     'MgOH+' 'NaOH+' 'OH-' 'OH+' 'SiH+' 'SrOH+']

    >>> # Only diatomic Na species:
    >>> diatomic = janaf.find_species(['Na'], num_atoms=2, charge='all')
    >>> print(diatomic)
    ['Na2' 'NaBr' 'NaCl' 'NaF' 'NaH' 'NaO' 'NaO-']
    """
    # Turn into dict if needed:
    if not isinstance(elements, dict):
        elements = {e:None for e in elements}

    janaf_dict = {}
    for line in open(f'{u.ROOT}chemcat/data/janaf_conversion.txt', 'r'):
        species_name, janaf_name = line.split()
        janaf_dict[species_name] = janaf_name

    species = []
    for molec, janaf_file in janaf_dict.items():
        if state == 'gas' and '_' in molec:
            continue

        is_charged = '+' in molec or '-' in molec
        if charge == 'neutral' and is_charged:
            continue
        elif charge == 'ion' and not is_charged:
            continue

        stoich = read_stoich(janaf_file=janaf_file)

        if num_atoms is not None:
            n_atoms = sum([val for key,val in stoich.items() if key!='e'])
            if n_atoms != num_atoms:
                continue

        # Request selected stoichiometric values:
        for key, val in elements.items():
            if key not in stoich or (val is not None and stoich[key] != val):
                break
        else:
            species.append(molec)

    return np.array(species)





# Copyright (c) 2022-2025 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

import numpy as np
import pytest

import chemcat.cea as cea



def test_is_in():
    species = 'H2O (KOH)2 HO2 CO'.split()
    in_cea = cea.is_in(species)
    np.testing.assert_equal(in_cea, np.array([True, False, True, True]))


def test_read_thermo_build():
    species = ['H2O']
    cea_data = cea.read_thermo_build(species)

    expected_t_coeffs = np.array([200.0, 1000.0, 6000.0, 0.0])
    expected_a_coeffs = np.array([
        [-3.94796083e+04,  5.75573102e+02,  9.31782653e-01,
          7.22271286e-03, -7.34255737e-06,  4.95504349e-09,
         -1.33693325e-12],
        [ 1.03497210e+06, -2.41269856e+03,  4.64611078e+00,
          2.29199831e-03, -6.83683048e-07,  9.42646893e-11,
         -4.82238053e-15],
        [ 0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
          0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
          0.00000000e+00]
    ])
    expected_b_coeffs = np.array([
        [-3.30397431e+04,  1.72420578e+01],
        [-1.38428651e+04, -7.97814851e+00],
        [ 0.00000000e+00,  0.00000000e+00]
    ])

    assert len(cea_data) == 1
    data = cea_data[0]
    assert data['name'] == 'H2O'
    assert data['stoich'] == {'H':2.0, 'O':1.0}

    assert np.shape(data['t_coeffs']) == (4,)
    assert np.shape(data['a_coeffs']) == (3,7)
    assert np.shape(data['b_coeffs']) == (3,2)
    np.testing.assert_allclose(data['t_coeffs'], expected_t_coeffs)
    np.testing.assert_allclose(data['a_coeffs'], expected_a_coeffs)
    np.testing.assert_allclose(data['b_coeffs'], expected_b_coeffs)


def test_read_thermo_build_stoich_neutral():
    species = 'H H2 H2O'.split()
    cea_data = cea.read_thermo_build(species)

    stoich = cea_data[0]['stoich']
    assert len(stoich) == 1
    assert stoich['H'] == 1.0

    stoich = cea_data[1]['stoich']
    assert len(stoich) == 1
    assert stoich['H'] == 2.0

    stoich = cea_data[2]['stoich']
    assert len(stoich) == 2
    assert stoich['H'] == 2.0
    assert stoich['O'] == 1.0


def test_read_thermo_cea_stoich_ions():
    species = 'e- H- H+ H3O+'.split()
    cea_data = cea.read_thermo_build(species)

    stoich = cea_data[0]['stoich']
    assert len(stoich) == 1
    assert stoich['e'] == 1.0

    stoich = cea_data[1]['stoich']
    assert len(stoich) == 2
    assert stoich['H'] == 1.0
    assert stoich['e'] == 1.0

    stoich = cea_data[2]['stoich']
    assert len(stoich) == 2
    assert stoich['H'] == 1.0
    assert stoich['e'] == -1.0

    stoich = cea_data[3]['stoich']
    assert len(stoich) == 3
    assert stoich['H'] == 3.0
    assert stoich['O'] == 1.0
    assert stoich['e'] == -1.0


@pytest.mark.parametrize(
    'temp',
    [300, 300.0],
)
def test_heat_func_cea_single_value(temp):
    data = cea.read_thermo_build(['H2O'])[0]
    heat = cea.heat_func(data['a_coeffs'], data['t_coeffs'])
    np.testing.assert_allclose(heat(temp), np.array([4.04063805]))


@pytest.mark.parametrize(
    'temp',
    ([300, 1000.0, 3000.0], np.array([300, 1000.0, 3000.0]))
)
def test_heat_func_cea_array(temp):
    data = cea.read_thermo_build(['H2O'])[0]
    heat = cea.heat_func(data['a_coeffs'], data['t_coeffs'])
    expected_heat = np.array([4.04063805, 4.96614188, 6.8342561])
    np.testing.assert_allclose(heat(temp), expected_heat)


def test_setup_network_cea_neutrals():
    molecules = 'H2O CH4 CO2 H2 H He'.split()
    cea_data = cea.setup_network(molecules)

    expected_stoich = [
        {'H': 2.0, 'O': 1.0},
        {'C': 1.0, 'H': 4.0},
        {'C': 1.0, 'O': 2.0},
        {'H': 2.0},
        {'H': 1.0},
        {'He': 1.0},
    ]

    assert len(cea_data) == 4
    np.testing.assert_equal(cea_data[0], molecules)
    np.testing.assert_equal(cea_data[3], expected_stoich)


def test_setup_network_cea_ions():
    molecules = 'H2O CH4 CO2 H2 H He e- H- H2+'.split()
    cea_data = cea.setup_network(molecules)

    expected_stoich = [
        {'H': 2.0, 'O': 1.0},
        {'C': 1.0, 'H': 4.0},
        {'C': 1.0, 'O': 2.0},
        {'H': 2.0},
        {'H': 1.0},
        {'He': 1.0},
        {'e': 1.0},
        {'H': 1.0, 'e': 1.0},
        {'H': 2.0, 'e': -1.0},
    ]

    np.testing.assert_equal(cea_data[0], molecules)
    np.testing.assert_equal(cea_data[3], expected_stoich)


def test_setup_network_cea_missing_species():
    molecules = 'Ti Ti+ TiO TiO2+ TiO2'.split()
    cea_data = cea.setup_network(molecules)

    expected_stoich= [
        {'Ti': 1.0},
        {'Ti': 1.0, 'e': -1.0},
        {'Ti': 1.0, 'O': 1.0},
        {'Ti': 1.0, 'O': 2.0},
    ]

    np.testing.assert_equal(cea_data[0], ['Ti', 'Ti+', 'TiO', 'TiO2'])
    np.testing.assert_equal(cea_data[3], expected_stoich)


def test_find_species_single():
    specs = cea.find_species(['K'])
    assert list(specs) == [
       'K', 'KCN', 'KH', 'KNO2', 'KNO3', 'KNa', 'KO', 'KOH', 'K2',
       'K2CO3', 'K2C2N2', 'K2O', 'K2O2', 'K2O2H2']


def test_find_species_dict_without_stoich_values():
    specs = cea.find_species({'K':None})
    assert list(specs) == [
       'K', 'KCN', 'KH', 'KNO2', 'KNO3', 'KNa', 'KO', 'KOH', 'K2',
       'K2CO3', 'K2C2N2', 'K2O', 'K2O2', 'K2O2H2']


def test_find_species_with_stoich_values():
    specs = cea.find_species({'K':2})
    assert list(specs) == ['K2', 'K2CO3', 'K2C2N2', 'K2O', 'K2O2', 'K2O2H2']


def test_find_species_multiple():
    specs = cea.find_species('N O'.split())
    assert list(specs) == [
        'OCCN', 'CNCOCN', 'HNCO', 'HNO', 'HNO2', 'HNO3', 'KNO2', 'KNO3',
        'NCO', 'NH2OH', 'NO', 'NO2', 'NO3', 'N2O', 'NH2NO2', 'N2O3',
        'N2O4', 'N2O5', 'NaNO2', 'NaNO3']


def test_find_species_ions():
    specs = cea.find_species('N O'.split(), charge='ion')
    assert list(specs) == ['NO+', 'NO2-', 'NO3-', 'N2O+']


def test_find_species_neutral_and_ions():
    specs = cea.find_species('N O'.split(), charge='all')
    assert list(specs) == [
        'OCCN', 'CNCOCN', 'HNCO', 'HNO', 'HNO2', 'HNO3', 'KNO2', 'KNO3',
        'NCO', 'NH2OH', 'NO', 'NO+', 'NO2', 'NO2-', 'NO3', 'NO3-', 'N2O',
        'NH2NO2', 'N2O+', 'N2O3', 'N2O4', 'N2O5', 'NaNO2', 'NaNO3',
    ]


def test_find_species_num_atoms():
    specs = cea.find_species(['Na'], num_atoms=2, charge='all')
    assert list(specs) == ['KNa', 'NaH', 'NaO', 'Na2']



# Copyright (c) 2022-2025 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

import os
import re
import setuptools
from setuptools import setup, Extension

from numpy import get_include


srcdir = 'src_c/'  # C-code source folder
incdir = 'src_c/include/'  # Include filder with header files

cfiles = os.listdir(srcdir)
cfiles = list(filter(lambda x: re.search('.+[.]c$', x), cfiles))
cfiles = list(filter(lambda x: not re.search('[.#].+[.]c$', x), cfiles))

inc = [get_include(), incdir]
eca = ['-lm', '-O3', '-ffast-math']
ela = ['-lm']

extensions = [
    Extension(
        'chemcat.lib.' + cfile.rstrip('.c'),
        sources=[f'{srcdir}{cfile}'],
        include_dirs=inc,
        extra_compile_args=eca,
        extra_link_args=ela,
    )
    for cfile in cfiles
]

setup(
    include_dirs = inc,
    ext_modules = extensions,
)




# Copyright (c) 2022-2025 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

import os

import numpy as np
import pytest

import chemcat as cat
import chemcat.janaf as janaf
import chemcat.utils as u

from conftest import *


# Some default values:
nlayers = 11
net_temperature = np.tile(1200.0, nlayers)
net_pressure = np.logspace(-8, 3, nlayers)
net_molecules = 'H2O CH4 CO CO2 NH3 N2 H2 HCN OH H He C N O'.split()


def test_network_init():
    nlayers = 81
    temperature = np.tile(1200.0, nlayers)
    pressure = np.logspace(-8, 3, nlayers)
    molecules = 'H2O CH4 CO CO2 H2 H C O'.split()
    net = cat.Network(pressure, temperature, molecules)

    expected_stoich_vals = np.array([
        [0, 2, 1],
        [1, 4, 0],
        [1, 0, 1],
        [1, 0, 2],
        [0, 2, 0],
        [0, 1, 0],
        [1, 0, 0],
        [0, 0, 1]])
    expected_element_rel_abundance = [2.88403150e-04, 1.0, 4.89778819e-04]
    expected_provenance = np.array([
        'janaf', 'janaf', 'janaf', 'janaf', 'janaf', 'janaf', 'janaf', 'janaf',
    ])
    expected_element_file = \
        f'{u.ROOT}chemcat/data/asplund_2021_solar_abundances.dat'

    np.testing.assert_equal(net.pressure, pressure)
    np.testing.assert_equal(net.temperature, temperature)
    np.testing.assert_equal(net.input_species, molecules)
    assert net.metallicity == 0.0
    assert net.e_abundances == {}
    assert net.e_scale == {}
    assert net.e_ratio == {}
    assert net.element_file == expected_element_file

    np.testing.assert_equal(net.species, molecules)
    np.testing.assert_equal(net.elements, ['C', 'H', 'O'])
    np.testing.assert_equal(net.provenance, expected_provenance)
    np.testing.assert_equal(net.stoich_vals, expected_stoich_vals)
    np.testing.assert_allclose(
        net.element_rel_abundance, expected_element_rel_abundance,
    )


def test_network_mixed_sources():
    # At least two things are being checked here:
    # - COOH only exists in cea,
    # - H2O2 is named HOOH in janaf,
    nlayers = 81
    temperature = np.tile(1200.0, nlayers)
    pressure = np.logspace(-8, 3, nlayers)
    molecules = ['H2O', 'H2O2', 'COOH', 'H']
    sources = 'janaf', 'cea'
    net = cat.Network(
        pressure, temperature, molecules, sources=sources,
    )

    expected_stoich_vals = np.array([
        [0, 2, 1],
        [0, 2, 2],
        [1, 1, 2],
        [0, 1, 0]])
    expected_provenance = np.array(['janaf', 'janaf', 'cea', 'janaf'])
    expected_species = ['H2O', 'H2O2', 'COOH', 'H']

    np.testing.assert_equal(net.pressure, pressure)
    np.testing.assert_equal(net.temperature, temperature)
    np.testing.assert_equal(net.input_species, molecules)
    np.testing.assert_equal(net.species, expected_species)
    np.testing.assert_equal(net.provenance, expected_provenance)
    np.testing.assert_equal(net.elements, ['C', 'H', 'O'])
    np.testing.assert_equal(net.stoich_vals, expected_stoich_vals)


def test_network_missing_species(capfd):
    nlayers = 81
    temperature = np.tile(1200.0, nlayers)
    pressure = np.logspace(-8, 3, nlayers)
    molecules = ['H2O', 'H2O2', 'FeH', 'COOH', 'H']
    net = cat.Network(pressure, temperature, molecules)

    expected_stoich_vals = np.array([
        [0, 2, 1],
        [0, 2, 2],
        [1, 1, 2],
        [0, 1, 0]])
    expected_provenance = np.array(['janaf', 'janaf', 'cea', 'janaf'])
    expected_species = ['H2O', 'H2O2', 'COOH', 'H']

    captured = capfd.readouterr()
    assert 'input species were not found in any database' in captured.out
    assert 'FeH' in captured.out

    np.testing.assert_equal(net.pressure, pressure)
    np.testing.assert_equal(net.temperature, temperature)
    np.testing.assert_equal(net.input_species, molecules)
    np.testing.assert_equal(net.species, expected_species)
    np.testing.assert_equal(net.provenance, expected_provenance)
    np.testing.assert_equal(net.elements, ['C', 'H', 'O'])
    np.testing.assert_equal(net.stoich_vals, expected_stoich_vals)


def test_network_cp_default_temp():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    cp = net.heat_capacity()

    expected_cp = np.array([
        5.26408044, 9.48143057, 4.11030773, 6.77638503, 7.34238673,
        4.05594463, 3.72748083, 6.3275286 , 3.79892261, 2.49998117,
        2.49998117, 2.50082308, 2.49998117, 2.51092596])

    assert np.shape(cp) == (len(net_pressure), len(net_molecules))
    np.testing.assert_allclose(cp[0], expected_cp)
    np.testing.assert_equal(net.temperature, net_temperature)


def test_network_duplicated_species(capfd):
    nlayers = 81
    temperature = np.tile(1200.0, nlayers)
    pressure = np.logspace(-8, 3, nlayers)
    molecules = ['H', 'C', 'O', 'H2O', 'H2', 'CO', 'H2O', 'CO2']
    net = cat.Network(pressure, temperature, molecules)
    vmr = net.thermochemical_equilibrium()

    good_molecules = ['H', 'C', 'O', 'H2O', 'H2', 'CO', 'CO2']
    good_net = cat.Network(pressure, temperature, good_molecules)
    expected_vmr = good_net.thermochemical_equilibrium()

    expected_stoich_vals = np.array([
        [0, 1, 0],
        [1, 0, 0],
        [0, 0, 1],
        [0, 2, 1],
        [0, 2, 0],
        [1, 0, 1],
        [1, 0, 2],
    ])
    expected_provenance = [
        'janaf', 'janaf', 'janaf', 'janaf', 'janaf', 'janaf', 'janaf',
    ]
    expected_species = ['H', 'C', 'O', 'H2O', 'H2', 'CO', 'CO2']

    captured = capfd.readouterr()
    assert 'These species are duplicates of others in input' in captured.out
    assert 'H2O' in captured.out

    np.testing.assert_equal(net.pressure, pressure)
    np.testing.assert_equal(net.temperature, temperature)
    np.testing.assert_equal(net.input_species, molecules)
    np.testing.assert_equal(net.species, expected_species)
    np.testing.assert_equal(net.provenance, expected_provenance)
    np.testing.assert_equal(net.elements, ['C', 'H', 'O'])
    np.testing.assert_equal(net.stoich_vals, expected_stoich_vals)
    np.testing.assert_equal(net.vmr, expected_vmr)


def test_network_duplicated_aliased_species(capfd):
    nlayers = 81
    temperature = np.tile(1200.0, nlayers)
    pressure = np.logspace(-8, 3, nlayers)
    molecules = ['H', 'S',  'H2O', 'SH', 'H2S', 'SH']
    net = cat.Network(pressure, temperature, molecules)
    vmr = net.thermochemical_equilibrium()

    good_molecules = ['H', 'S', 'H2O', 'SH', 'H2S']
    good_net = cat.Network(pressure, temperature, good_molecules)
    expected_vmr = good_net.thermochemical_equilibrium()

    expected_stoich_vals = np.array([
        [1, 0, 0],
        [0, 0, 1],
        [2, 1, 0],
        [1, 0, 1],
        [2, 0, 1],
    ])
    expected_provenance = ['janaf', 'janaf', 'janaf', 'janaf', 'janaf']
    expected_species = ['H', 'S', 'H2O', 'SH', 'H2S']

    captured = capfd.readouterr()
    assert 'These species are duplicates of others in input' in captured.out
    assert 'SH' in captured.out

    np.testing.assert_equal(net.pressure, pressure)
    np.testing.assert_equal(net.temperature, temperature)
    np.testing.assert_equal(net.input_species, molecules)
    np.testing.assert_equal(net.species, expected_species)
    np.testing.assert_equal(net.provenance, expected_provenance)
    np.testing.assert_equal(net.elements, ['H', 'O', 'S'])
    np.testing.assert_equal(net.stoich_vals, expected_stoich_vals)
    np.testing.assert_equal(net.vmr, expected_vmr)


def test_network_cp_input_temp():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    temps = [100.0, 600.0, 1200.0]
    cp = net.heat_capacity(temps)

    expected_cp = np.array([
        [4.00494915, 4.3688933 , 5.26408044],
        [4.00001798, 6.28146429, 9.48143057],
        [3.50040662, 3.6614513 , 4.11030773],
        [3.51291495, 5.69140811, 6.77638503],
        [4.00314507, 5.44749578, 7.34238673],
        [3.50040662, 3.62140061, 4.05594463],
        [3.38614788, 3.52722736, 3.72748083],
        [3.50798378, 5.26865079, 6.3275286 ],
        [3.92412613, 3.55128183, 3.79892261],
        [2.49998117, 2.49998117, 2.49998117],
        [2.49998117, 2.49998117, 2.49998117],
        [2.55831326, 2.50154471, 2.50082308],
        [2.49998117, 2.49998117, 2.49998117],
        [2.85081563, 2.54063323, 2.51092596]]).T

    assert np.shape(cp) == (len(temps), len(net_molecules))
    np.testing.assert_allclose(cp, expected_cp)
    # Network's temperature is not updated:
    np.testing.assert_equal(net.temperature, net_temperature)


def test_network_gibbs_default_temp():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    gibbs = net.gibbs_free_energy()

    expected_gibbs = np.array([
        -49.70275118, -33.59076581, -37.17544326, -68.58699428,
        -31.08382889, -25.35365299, -17.97578591, -13.94856633,
        -20.48067821,   6.44982554, -16.77498672,  51.21052551,
         27.33544072,   3.95818325])

    assert np.shape(gibbs) == (len(net_temperature), len(net_molecules))
    np.testing.assert_allclose(gibbs[0], expected_gibbs)
    np.testing.assert_equal(net.temperature, net_temperature)


def test_network_vmr_e_source_asplund():
    net = cat.Network(
        net_pressure, net_temperature, net_molecules,
        e_source='asplund_2021',
    )
    vmr = net.thermochemical_equilibrium()

    assert np.shape(vmr) == (len(net_temperature), len(net_molecules))
    np.testing.assert_allclose(vmr, expected_vmr_1200K)
    np.testing.assert_allclose(net.vmr, expected_vmr_1200K)

    expected_e_abundance = np.array(
        [2.88403150e-04, 1.0, 8.20351544e-02, 6.76082975e-05, 4.89778819e-04]
    )
    np.testing.assert_allclose(net.element_rel_abundance, expected_e_abundance)

    assert np.all(vmr>=0)
    elem_fractions = np.sum(net.vmr[0]*net.stoich_vals.T, axis=1)
    elem_fractions /= elem_fractions[net.elements == 'H']
    np.testing.assert_allclose(elem_fractions, net.element_rel_abundance)


def test_network_vmr_cea_priority():
    net = cat.Network(
        net_pressure, net_temperature, net_molecules,
        sources=['cea', 'janaf']
    )
    vmr = net.thermochemical_equilibrium()

    assert np.shape(vmr) == (len(net_temperature), len(net_molecules))
    np.testing.assert_allclose(net.vmr, expected_vmr_1200K_cea)

    expected_e_abundance = np.array(
        [2.88403150e-04, 1.0, 8.20351544e-02, 6.76082975e-05, 4.89778819e-04]
    )
    np.testing.assert_allclose(net.element_rel_abundance, expected_e_abundance)

    assert np.all(vmr>=0)
    elem_fractions = np.sum(net.vmr[0]*net.stoich_vals.T, axis=1)
    elem_fractions /= elem_fractions[net.elements == 'H']
    np.testing.assert_allclose(elem_fractions, net.element_rel_abundance)


def test_network_vmr_update_temp():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    vmr = net.thermochemical_equilibrium()

    assert np.shape(vmr) == (len(net_temperature), len(net_molecules))
    np.testing.assert_allclose(vmr, expected_vmr_1200K)

    thot = np.tile(2900.0, nlayers)
    vmr = net.thermochemical_equilibrium(temperature=thot)
    assert np.shape(vmr) == (len(net_temperature), len(net_molecules))
    np.testing.assert_allclose(vmr, expected_vmr_2900K)
    np.testing.assert_equal(net.temperature, thot)



def test_network_vmr_update_metallicity():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    vmr = net.thermochemical_equilibrium(metallicity=0.0)
    np.testing.assert_allclose(vmr, expected_vmr_1200K)

    vmr = net.thermochemical_equilibrium(metallicity=1.0)
    expected_e_abundance = np.array(
        [2.88403150e-03, 1.0, 8.20351544e-02, 6.76082975e-04, 4.89778819e-03]
    )
    np.testing.assert_allclose(vmr, expected_vmr_10x_solar)
    np.testing.assert_equal(net.metallicity, 1.0)
    np.testing.assert_allclose(net.element_rel_abundance, expected_e_abundance)

    vmr = net.thermochemical_equilibrium()
    np.testing.assert_allclose(vmr, expected_vmr_10x_solar)


def test_network_vmr_update_e_abundances():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    e_abundances = {'C': 8.0}
    vmr = net.thermochemical_equilibrium(e_abundances=e_abundances)

    np.testing.assert_allclose(vmr, expected_vmr_e_abund)
    expected_e_abundance = np.array(
        [1.0e-04, 1.0, 8.20351544e-02, 6.76082975e-05, 4.89778819e-04]
    )
    np.testing.assert_allclose(net.element_rel_abundance, expected_e_abundance)


def test_network_vmr_update_e_scale():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    e_scale = {'C': 1.0}
    vmr = net.thermochemical_equilibrium(e_scale=e_scale)

    np.testing.assert_allclose(vmr, expected_vmr_e_scale)
    expected_e_abundance = np.array(
        [2.88403150e-03, 1.0, 8.20351544e-02, 6.76082975e-05, 4.89778819e-04]
    )
    np.testing.assert_allclose(net.element_rel_abundance, expected_e_abundance)


def test_network_vmr_update_e_ratio():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    e_ratio = {'C_O': 0.8}
    vmr = net.thermochemical_equilibrium(e_ratio=e_ratio)

    np.testing.assert_allclose(vmr, expected_vmr_e_ratio)
    expected_e_abundance = np.array(
        [3.91823055e-04, 1.0, 8.20351544e-02, 6.76082975e-05, 4.89778819e-04]
    )
    np.testing.assert_allclose(net.element_rel_abundance, expected_e_abundance)
    C_to_O = (
        net.element_rel_abundance[net.elements=='C']
        / net.element_rel_abundance[net.elements=='O'])
    np.testing.assert_allclose(C_to_O, 0.8)


def test_network_vmr_update_metal_e_ratio():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    metallicity = 1.0
    e_ratio = {'C_O': 1.1}
    vmr = net.thermochemical_equilibrium(
        metallicity=metallicity,
        e_ratio=e_ratio,
    )

    np.testing.assert_allclose(vmr, expected_vmr_metal_e_ratio)
    expected_e_abundance = np.array(
        [5.38756701e-03, 1.0, 8.20351544e-02, 6.76082975e-04, 4.89778819e-03]
    )
    np.testing.assert_allclose(
        net.element_rel_abundance, expected_e_abundance)
    C_to_O = (
        net.element_rel_abundance[net.elements=='C']
        / net.element_rel_abundance[net.elements=='O'])
    np.testing.assert_allclose(C_to_O, 1.1)


def test_network_vmr_update_temp_error():
    net = cat.Network(net_pressure, net_temperature, net_molecules)
    match = (
        'Temperature profile does not match size of '
        'pressure profile [(]11 layers[)]')
    bad_temp = np.tile(1200.0, nlayers-1)
    with pytest.raises(ValueError, match=match):
        net.thermochemical_equilibrium(bad_temp)


def test_network_vmr_write_file(tmpdir):
    atm_file = "network_atm_file.dat"
    atm = f'{tmpdir}/{atm_file}'

    net = cat.Network(net_pressure, net_temperature, net_molecules)
    vmr = net.thermochemical_equilibrium(savefile=atm)
    assert atm_file in os.listdir(str(tmpdir))



